# @license
# Copyright 2025 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Writes annotations in the Precomputed annotation format.

This provides a simple way to write annotations in the precomputed format, but
has a number of limitations that makes it suitable only for a relatively small
amount of annotation data:

- All annotations are buffered in memory.

- Only a trivial spatial index consisting of a single grid cell at a single
  level is generated.  Consequently, Neuroglancer will be forced to download all
  annotations at once.

- All indices are written in the unsharded format.  Consequently, there is at
  least one file written per annotation.
"""

import json
import numbers
import os
import pathlib
import struct
from collections.abc import Sequence
from typing import Literal, NamedTuple, cast

import numpy as np

from . import coordinate_space, viewer_state


class Annotation(NamedTuple):
    id: int
    encoded: bytes
    relationships: Sequence[Sequence[int]]


_PROPERTY_DTYPES: dict[str, tuple[tuple[str] | tuple[str, tuple[int, ...]], int]] = {
    "uint8": (("|u1",), 1),
    "uint16": (("<u2",), 2),
    "uint32": (("<u4",), 3),
    "int8": (("|i1",), 1),
    "int16": (("<i2",), 2),
    "int32": (("<i4",), 4),
    "float32": (("<f4",), 4),
    "rgb": (("|u1", (3,)), 1),
    "rgba": (("|u1", (4,)), 1),
}

AnnotationType = Literal["point", "line", "axis_aligned_bounding_box", "ellipsoid"]


def _get_dtype_for_geometry(annotation_type: AnnotationType, rank: int):
    geometry_size = rank if annotation_type == "point" else 2 * rank
    return [("geometry", "<f4", geometry_size)]


def _get_dtype_for_properties(
    properties: Sequence[viewer_state.AnnotationPropertySpec],
):
    dtype = []
    offset = 0
    for i, p in enumerate(properties):
        dtype_entry, alignment = _PROPERTY_DTYPES[p.type]
        if offset % alignment:
            padded_offset = (offset + alignment - 1) // alignment * alignment
            padding = padded_offset - offset
            dtype.append((f"padding{offset}", "|u1", (padding,)))
            offset += padding
        dtype.append((f"property{i}", *dtype_entry))  # type: ignore[arg-type]
        size = np.dtype(dtype[-1:]).itemsize
        offset += size
    alignment = 4
    if offset % alignment:
        padded_offset = (offset + alignment - 1) // alignment * alignment
        padding = padded_offset - offset
        dtype.append((f"padding{offset}", "|u1", (padding,)))
        offset += padding
    return dtype


def _convert_rgb_to_uint8(rgb: str) -> tuple[int, int, int]:
    """Convert an RGB hex string to a tuple of uint8 values."""
    if rgb.startswith("#"):
        rgb = rgb[1:]
    if len(rgb) != 6:
        raise ValueError(f"Invalid RGB format: {rgb}")
    return (int(rgb[0:2], 16), int(rgb[2:4], 16), int(rgb[4:6], 16))


def _convert_rgba_to_uint8(rgba: str) -> tuple[int, int, int, int]:
    """Convert an RGBA hex string to a tuple of uint8 values."""
    if rgba.startswith("#"):
        rgba = rgba[1:]
    if len(rgba) != 8:
        raise ValueError(f"Invalid RGBA format: {rgba}")
    color = _convert_rgb_to_uint8(rgba[:6])
    alpha = int(rgba[6:8], 16)
    return (*color, alpha)


class AnnotationWriter:
    annotations: list[Annotation]
    related_annotations: list[dict[int, list[Annotation]]]
    lower_bound: np.typing.NDArray[np.float64]
    upper_bound: np.typing.NDArray[np.float64]

    def __init__(
        self,
        coordinate_space: coordinate_space.CoordinateSpace,
        annotation_type: AnnotationType,
        relationships: Sequence[str] = (),
        properties: Sequence[viewer_state.AnnotationPropertySpec] = (),
    ):
        self.coordinate_space = coordinate_space
        self.relationships = list(relationships)
        self.annotation_type = annotation_type
        self.properties = list(properties)
        self.properties.sort(key=lambda p: -_PROPERTY_DTYPES[p.type][1])
        self.annotations = []
        self.rank = coordinate_space.rank
        self.dtype = _get_dtype_for_geometry(
            annotation_type, coordinate_space.rank
        ) + _get_dtype_for_properties(self.properties)
        self.lower_bound = np.full(
            shape=(self.rank,), fill_value=float("inf"), dtype=np.float32
        )
        self.upper_bound = np.full(
            shape=(self.rank,), fill_value=float("-inf"), dtype=np.float32
        )
        self.related_annotations = [{} for _ in self.relationships]

    def add_point(self, point: Sequence[float], id: int | None = None, **kwargs):
        if self.annotation_type != "point":
            raise ValueError(
                f"Expected annotation type point, but received: {self.annotation_type}"
            )
        if len(point) != self.coordinate_space.rank:
            raise ValueError(
                f"Expected point to have length {self.coordinate_space.rank}, but received: {len(point)}"
            )

        self.lower_bound = np.minimum(self.lower_bound, point)
        self.upper_bound = np.maximum(self.upper_bound, point)
        self._add_obj(point, id, **kwargs)

    def add_axis_aligned_bounding_box(
        self,
        point_a: Sequence[float],
        point_b: Sequence[float],
        id: int | None = None,
        **kwargs,
    ):
        if self.annotation_type != "axis_aligned_bounding_box":
            raise ValueError(
                f"Expected annotation type axis_aligned_bounding_box, but received: {self.annotation_type}"
            )
        self._add_two_point_obj(point_a, point_b, id, **kwargs)

    def add_line(
        self,
        point_a: Sequence[float],
        point_b: Sequence[float],
        id: int | None = None,
        **kwargs,
    ):
        if self.annotation_type != "line":
            raise ValueError(
                f"Expected annotation type line, but received: {self.annotation_type}"
            )
        self._add_two_point_obj(point_a, point_b, id, **kwargs)

    def _add_two_point_obj(
        self,
        point_a: Sequence[float],
        point_b: Sequence[float],
        id: int | None = None,
        **kwargs,
    ):
        if len(point_a) != self.coordinate_space.rank:
            raise ValueError(
                f"Expected coordinates to have length {self.coordinate_space.rank}, but received: {len(point_a)}"
            )

        if len(point_b) != self.coordinate_space.rank:
            raise ValueError(
                f"Expected coordinates to have length {self.coordinate_space.rank}, but received: {len(point_b)}"
            )

        self.lower_bound = np.minimum(self.lower_bound, point_a)
        self.upper_bound = np.maximum(self.upper_bound, point_b)
        coords = np.concatenate((point_a, point_b))
        self._add_obj(cast(Sequence[float], coords), id, **kwargs)

    def _add_obj(self, coords: Sequence[float], id: int | None, **kwargs):
        encoded = np.zeros(shape=(), dtype=self.dtype)
        encoded[()]["geometry"] = coords  # type: ignore[call-overload]

        for i, p in enumerate(self.properties):
            default_value = p.default
            if p.id in kwargs:
                default_value = kwargs.pop(p.id)
            if isinstance(default_value, str) and p.type in ("rgb", "rgba"):
                if p.type == "rgb":
                    default_value = _convert_rgb_to_uint8(default_value)
                else:
                    default_value = _convert_rgba_to_uint8(default_value)
            if default_value is not None:
                encoded[()][f"property{i}"] = default_value  # type: ignore[call-overload]

        related_ids = []
        for relationship in self.relationships:
            ids = kwargs.pop(relationship, None)
            if ids is None:
                ids = []
            if isinstance(ids, numbers.Integral):
                ids = [ids]
            related_ids.append(ids)

        if kwargs:
            raise ValueError(f"Unexpected keyword arguments {kwargs}")

        if id is None:
            id = len(self.annotations)

        annotation = Annotation(
            id=id, encoded=encoded.tobytes(), relationships=related_ids
        )

        self.annotations.append(annotation)

        for i, segment_ids in enumerate(related_ids):
            for segment_id in segment_ids:
                rel_index = self.related_annotations[i]
                rel_index_list = rel_index.setdefault(segment_id, [])
                rel_index_list.append(annotation)

    def _serialize_annotations(self, f, annotations: list[Annotation]):
        f.write(struct.pack("<Q", len(annotations)))
        for annotation in annotations:
            f.write(annotation.encoded)
        for annotation in annotations:
            f.write(struct.pack("<Q", annotation.id))

    def _serialize_annotation(self, f, annotation: Annotation):
        f.write(annotation.encoded)
        for related_ids in annotation.relationships:
            f.write(struct.pack("<I", len(related_ids)))
            for related_id in related_ids:
                f.write(struct.pack("<Q", related_id))

    def write(self, path: str | pathlib.Path):
        metadata = {
            "@type": "neuroglancer_annotations_v1",
            "dimensions": self.coordinate_space.to_json(),
            "lower_bound": [float(x) for x in self.lower_bound],
            "upper_bound": [float(x) for x in self.upper_bound],
            "annotation_type": self.annotation_type,
            "properties": [p.to_json() for p in self.properties],
            "relationships": [
                {"id": relationship, "key": f"rel_{relationship}"}
                for relationship in self.relationships
            ],
            "by_id": {
                "key": "by_id",
            },
            "spatial": [
                {
                    "key": "spatial0",
                    "grid_shape": [1] * self.rank,
                    "chunk_size": [
                        max(1, float(x)) for x in self.upper_bound - self.lower_bound
                    ],
                    "limit": len(self.annotations),
                },
            ],
        }

        os.makedirs(path, exist_ok=True)
        for relationship in self.relationships:
            os.makedirs(os.path.join(path, f"rel_{relationship}"), exist_ok=True)
        os.makedirs(os.path.join(path, "by_id"), exist_ok=True)
        os.makedirs(os.path.join(path, "spatial0"), exist_ok=True)

        with open(os.path.join(path, "info"), "w") as f:
            f.write(json.dumps(metadata))

        with open(
            os.path.join(path, "spatial0", "_".join("0" for _ in range(self.rank))),
            "wb",
        ) as f:
            self._serialize_annotations(f, self.annotations)

        for annotation in self.annotations:
            with open(os.path.join(path, "by_id", str(annotation.id)), "wb") as f:
                self._serialize_annotation(f, annotation)

        for i, relationship in enumerate(self.relationships):
            rel_index = self.related_annotations[i]
            for segment_id, annotations in rel_index.items():
                with open(
                    os.path.join(path, f"rel_{relationship}", str(segment_id)), "wb"
                ) as f:
                    self._serialize_annotations(f, annotations)



"""Entrypoint to a local neuroglancer instance and associated shell."""
# stdlib
import argparse
import logging
import sys
import webbrowser
from functools import partial

# internals
from ngtools.local.fileserver import LocalFileServer
from ngtools.local.termcolors import bformat
from ngtools.local.viewer import LocalNeuroglancer

LOGO = r"""
             _              _
 _ __   __ _| |_ ___   ___ | |___
| '_ \ / _` | __/ _ \ / _ \| / __|
| | | | (_| | || (_) | (_) | \__ \
|_| |_|\__, |\__\___/ \___/|_|___/
       |___/
"""


def main(args: list[str] | None = None) -> None:
    """Commandline launcher for local Neuroglancer instances."""
    args = args or sys.argv[1:]

    help_fmt = partial(argparse.HelpFormatter, max_help_position=32)

    log_choices = ("any", "debug", "info", "warning", "error", "none")
    log_map = {
        "a": 0,
        "d": logging.DEBUG,
        "i": logging.INFO,
        "w": logging.WARNING,
        "e": logging.ERROR,
        "n": 1024
    }

    parser = argparse.ArgumentParser('Run a local neuroglancer',
                                     formatter_class=help_fmt)
    parser.add_argument('--token', type=str, default='1',
                        help="neuroglancer unique token")
    parser.add_argument('--ip', default='127.0.0.1',
                        help="local IP")
    parser.add_argument('--port', type=int, default=9321,
                        help="viewer port", metavar="PORT")
    parser.add_argument('--port-fileserver', type=int, default=9123,
                        help="fileserver port", metavar="PORT")
    parser.add_argument('--no-window', action='store_true', default=False,
                        help="do not open neuroglancer window")
    parser.add_argument('--no-fileserver', action='store_true', default=False,
                        help="do not run a local fileserver")
    parser.add_argument('--debug', action='store_true', default=False,
                        help="run in debug mode")
    parser.add_argument('--log-level', choices=log_choices, default=None,
                        help="logging level")
    parser.add_argument('--stdin', help="input stream (default: stdin)")
    parser.add_argument('--stdout', help="output stream (default: stdout)")
    parser.add_argument('--stderr', help="error stream (default: stderr)")
    parser.add_argument(nargs='*', dest='filenames', help='Files to load')
    args = parser.parse_args(args)

    if args.ip == 'auto':
        args.ip = ''

    # -------
    # Logging
    # -------

    if args.log_level is None:
        args.log_level = "debug" if args.debug else "none"
    args.log_level = log_map[args.log_level[0].lower()]

    logging.basicConfig()
    logging.getLogger().setLevel(1000)
    logging.getLogger("dandi").setLevel(1000)
    logging.getLogger("ngtools").setLevel(args.log_level)

    # ----------
    # Fileserver
    # ----------

    fileserver = False
    if not args.no_fileserver:
        fileserver = LocalFileServer(args.port_fileserver, args.ip)

    # ------
    # Viewer
    # ------

    neuroglancer = LocalNeuroglancer(
        port=args.port,
        ip=args.ip,
        token=args.token,
        fileserver=fileserver,
        debug=args.debug,
        stdin=args.stdin or sys.stdin,
        stdout=args.stdout or sys.stdout,
        stderr=args.stderr or sys.stderr,
    )
    url = neuroglancer.get_viewer_url()

    logo = bformat.bold(bformat.fg.blue(LOGO[1:-1]))
    neuroglancer.console.stdio.print(logo)
    neuroglancer.console.stdio.print('neuroglancer:', url)
    if neuroglancer.viewer:
        fileserver_url = neuroglancer.get_fileserver_url()
        neuroglancer.console.stdio.print('fileserver:  ', fileserver_url)

    if not args.no_window:
        webbrowser.open(url)

    # ----------
    # Load files
    # ----------

    import time
    time.sleep(1)
    for filename in (args.filenames or []):
        neuroglancer.load(filename)

    neuroglancer.await_input()

    # with neuroglancer.console.stdio.stdin as f:
    #     f.writelines([f'load "{filename}"' for filename in args.filenames])



import numpy as np
from nibabel.wrapstruct import LabeledWrapStruct as LWS


def shape_zoom_affine(shape, zooms, x_flip=True, y_flip=False):
    """
    Get affine implied by given shape and zooms.

    We get the translations from the center of the image (implied by
    `shape`).

    Parameters
    ----------
    shape : (N,) array-like
       shape of image data. ``N`` is the number of dimensions
    zooms : (N,) array-like
       zooms (voxel sizes) of the image
    x_flip : {True, False}
       whether to flip the X row of the affine.  Corresponds to
       radiological storage on disk.
    y_flip : {False, True}
       whether to flip the Y row of the affine.  Corresponds to
       DICOM storage on disk when x_flip is also True.

    Returns
    -------
    aff : (4,4) array
       affine giving correspondance of voxel coordinates to mm
       coordinates, taking the center of the image as origin

    Examples
    --------
    >>> shape = (3, 5, 7)
    >>> zooms = (3, 2, 1)
    >>> shape_zoom_affine((3, 5, 7), (3, 2, 1))
    array([[-3.,  0.,  0.,  3.],
           [ 0.,  2.,  0., -4.],
           [ 0.,  0.,  1., -3.],
           [ 0.,  0.,  0.,  1.]])
    >>> shape_zoom_affine((3, 5, 7), (3, 2, 1), False)
    array([[ 3.,  0.,  0., -3.],
           [ 0.,  2.,  0., -4.],
           [ 0.,  0.,  1., -3.],
           [ 0.,  0.,  0.,  1.]])

    """
    shape = np.asarray(shape)
    zooms = np.array(zooms)  # copy because of flip below
    ndims = len(shape)
    if ndims != len(zooms):
        raise ValueError("Should be same length of zooms and shape")
    if ndims >= 3:
        shape = shape[:3]
        zooms = zooms[:3]
    else:
        full_shape = np.ones((3,))
        full_zooms = np.ones((3,))
        full_shape[:ndims] = shape[:]
        full_zooms[:ndims] = zooms[:]
        shape = full_shape
        zooms = full_zooms
    if x_flip:
        zooms[0] *= -1

    if y_flip:
        zooms[1] *= -1
    # Get translations from center of image
    origin = (shape - 1) / 2.0
    aff = np.eye(4)
    aff[:3, :3] = np.diag(zooms)
    aff[:3, -1] = -origin * zooms
    return aff


class LabeledWrapStruct(LWS):
    def __setitem__(self, item, value):
        self._structarr[item] = np.asanyarray(value)



"""Lists of kown protocols."""

LAYERS = [
    'volume',           # Raster data (image or volume)
    'image',            # ^ alias
    'labels',           # Integer raster data, interpreted as labels
    'segmentation',     # ^ alias
    'surface',          # Triangular mesh
    'mesh',             # Other types of mesh ???
    'tracts',           # Set of piecewise curves
    'roi',              # Region of interest ???
    'points',           # Pointcloud
    'transform',        # Spatial transform
    'affine',           # Affine transform
]

NG_FORMATS = [
    'boss',             # bossDB: Block & Object storage system
    'brainmap',         # Google Brain Maps
    'deepzoom',         # Deep Zoom file-backed data source
    'dvid',             # DVID
    'graphene',         # Graphene Zoom file-backed data source
    'local',            # Local in-memory
    'n5',               # N5 data source
    'nggraph',          # nggraph data source
    'nifti',            # Single NIfTI file
    'obj',              # Wavefront OBJ mesh file
    'precomputed',      # Precomputed file-backed data source
    'render',           # Render
    'vtk',              # VTK mesh file
    'zarr',             # Zarr data source
    'zarr2',            # Zarr v2 data source
    'zarr3',            # Zarr v3 data source
]

# Extra local formats (if not specified, try to guess from file)
EXTRA_FORMATS = [
    'nibabel',          # All formats that ar eread by nibabel
    'mgh',              # Freesurfer volume format
    'mgz',              # Freesurfer volume format (compressed)
    'trk',              # Freesurfer streamlines
    'lta',              # Freesurfer affine transform
    'surf',             # Freesurfer surfaces
    'annot',            # Freesurfer surface annotation
    'tck',              # MRtrix streamlines
    'mif',              # MRtrix volume format
    'gii',              # Gifti
    'tiff',             # Tiff volume format
    'niftyreg',         # NiftyReg affine transform
]

FORMATS = NG_FORMATS + EXTRA_FORMATS

PROTOCOLS = [
    "file",          # Local file
    "http",          # HTTP
    "https",         # HTTPs
    "gs",            # Google Cloud
    "s3",            # Amazon S3
    "ftp",           # FTP
    "dandi",         # DANDI/LINC
    "local",         # protocol for LocalAnnotationLayer sources
    "python",        # protocol for LocalVolume/SkeletonSource
]



#!/usr/bin/env python
# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tool for saving object meshes from Neuroglancer using cloud-volume.

Usage:

python -m neuroglancer.tool.save_meshes --url 'https://neuroglancer-demo.appspot.com/#!%7B%22dimensions%22:%7B%22x%22:%5B8e-9%2C%22m%22%5D%2C%22y%22:%5B8e-9%2C%22m%22%5D%2C%22z%22:%5B8e-9%2C%22m%22%5D%7D%2C%22position%22:%5B2911.677978515625%2C3100.916748046875%2C4024.5%5D%2C%22crossSectionScale%22:3.762185354999915%2C%22projectionOrientation%22:%5B-0.4241213798522949%2C-0.04451639577746391%2C0.07846356928348541%2C-0.9011008739471436%5D%2C%22projectionScale%22:6078.433160286794%2C%22layers%22:%5B%7B%22type%22:%22image%22%2C%22source%22:%22precomputed://gs://neuroglancer-public-data/flyem_fib-25/image%22%2C%22name%22:%22image%22%7D%2C%7B%22type%22:%22segmentation%22%2C%22source%22:%22precomputed://gs://neuroglancer-public-data/flyem_fib-25/ground_truth%22%2C%22segments%22:%5B%2210319%22%2C%2224436%22%2C%222515%22%2C%2226353%22%2C%2250%22%5D%2C%22name%22:%22ground-truth%22%7D%5D%2C%22showSlices%22:false%2C%22layout%22:%224panel%22%2C%22partialViewport%22:%5B0%2C0%2C1%2C1%5D%7D' --output-dir mesh-output-dir

"""

import argparse
import os
import sys

import neuroglancer
import neuroglancer.cli

try:
    import cloudvolume
except ImportError:
    print("cloud-volume package is required: pip install cloud-volume")
    sys.exit(1)


def save_meshes(state, output_dir, output_format, lod):
    for layer in state.layers:
        if not isinstance(layer.layer, neuroglancer.SegmentationLayer):
            continue
        if not layer.visible:
            return False
        for source in layer.source:
            if not source.url.startswith("precomputed://"):
                continue
            vol = cloudvolume.CloudVolume(source.url, parallel=True, progress=True)
            if len(layer.segments) == 0:
                continue
            get_mesh_kwargs = {}
            if lod != 0:
                get_mesh_kwargs.update(lod=lod)
            for segment in layer.segments:
                output_path = os.path.join(
                    output_dir, "%d.%s" % (segment, output_format)
                )
                print(f"Saving layer {layer.name!r} object {segment} -> {output_path}")
                os.makedirs(output_dir, exist_ok=True)
                mesh = vol.mesh.get(segment, **get_mesh_kwargs)
                if isinstance(mesh, dict):
                    mesh = list(mesh.values())[0]
                if output_format == "obj":
                    data = mesh.to_obj()
                elif output_format == "ply":
                    data = mesh.to_ply()
                elif output_format == "precomputed":
                    data = mesh.to_precomputed()
                with open(output_path, "wb") as f:
                    f.write(data)
            return
    print("No segmentation layer found")
    sys.exit(1)


def main(args=None):
    ap = argparse.ArgumentParser()
    neuroglancer.cli.add_state_arguments(ap, required=True)
    ap.add_argument("--format", choices=["obj", "ply"], default="obj")
    ap.add_argument(
        "--lod", type=int, default=0, help="Mesh level of detail to download"
    )
    ap.add_argument("--output-dir", default=".")
    parsed_args = ap.parse_args()
    save_meshes(
        state=parsed_args.state,
        output_dir=parsed_args.output_dir,
        output_format=parsed_args.format,
        lod=parsed_args.lod,
    )


if __name__ == "__main__":
    main()



# @license
# Copyright 2023 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import threading

import tornado.platform.asyncio


class BackgroundTornadoServerMetaclass(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return obj


class BackgroundTornadoServer(metaclass=BackgroundTornadoServerMetaclass):
    def __init__(self, daemon=False):
        self._thread = threading.Thread(target=self._run_server)
        if daemon:
            self._thread.daemon = True
        self._started = concurrent.futures.Future()
        self._stop_lock = threading.Lock()
        self._stop_requested = False

    def __post_init__(self):
        self._thread.start()
        try:
            self._started.result()
        except:
            self._thread.join()
            raise

    def _run_server(self):
        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            tornado.platform.asyncio.AsyncIOMainLoop()
            self.loop = loop
            loop.call_soon(self._start_server)
            loop.run_forever()
            loop.close()
        except Exception as exc:
            self._started.set_exception(exc)

    def _attempt_to_start_server(self):
        raise NotImplementedError

    def _start_server(self):
        try:
            self._attempt_to_start_server()
            self._started.set_result(None)
        except Exception as e:
            self.loop.stop()
            self._started.set_exception(e)

    def request_stop(self):
        with self._stop_lock:
            if not self._stop_requested:
                self._stop_requested = True
                self.loop.call_soon_threadsafe(
                    lambda: asyncio.create_task(self._stop())
                )

    def stop(self):
        self.request_stop()
        self._thread.join()

    async def _stop(self):
        self.http_server.stop()
        await tornado.platform.asyncio.to_asyncio_future(
            self.http_server.close_all_connections()
        )
        self.loop.stop()



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import threading

import numpy as np

try:
    import tensorstore as ts
except ImportError:
    ts = None

from . import downsample, downsample_scales, trackable_state
from .chunks import encode_jpeg, encode_npz, encode_raw
from .coordinate_space import CoordinateSpace
from .random_token import make_random_token


class MeshImplementationNotAvailable(Exception):
    pass


class MeshesNotSupportedForVolume(Exception):
    pass


class InvalidObjectIdForMesh(Exception):
    pass


class LocalVolume(trackable_state.ChangeNotifier):
    """Exposes a local array as a Neuroglancer data source.

    Group:
      serve-data
    """

    def __init__(
        self,
        data,
        dimensions=None,
        volume_type=None,
        voxel_offset=None,
        encoding="npz",
        max_voxels_per_chunk_log2=None,
        mesh_options=None,
        downsampling="3d",
        chunk_layout=None,
        max_downsampling=downsample_scales.DEFAULT_MAX_DOWNSAMPLING,
        max_downsampled_size=downsample_scales.DEFAULT_MAX_DOWNSAMPLED_SIZE,
        max_downsampling_scales=downsample_scales.DEFAULT_MAX_DOWNSAMPLING_SCALES,
    ):
        """Initializes a LocalVolume.

        @param data: Source data.

        @param downsampling: '3d' to use isotropic downsampling, '2d' to
            downsample separately in XY, XZ, and YZ, None to use no
            downsampling.

        @param max_downsampling: Maximum amount by which on-the-fly downsampling
            may reduce the volume of a chunk.  For example, 4x4x4 downsampling
            reduces the volume by 64.

        @param volume_type: either 'image' or 'segmentation'.  If not specified,
            guessed from the data type.

        @param voxel_size: Sequence [x, y, z] of floats.  Specifies the voxel
            size.

        @param mesh_options: A dict with the following keys specifying options
            for mesh simplification for 'segmentation' volumes:

                - max_quadrics_error: float.  Edge collapses with a larger
                  associated quadrics error than this amount are prohibited.
                  Set this to a negative number to disable mesh simplification,
                  and just use the original mesh produced by the marching cubes
                  algorithm.  Defaults to 1e6.  The effect of this value depends
                  on the voxel_size.

                - max_normal_angle_deviation: float.  Edge collapses that change
                  a triangle normal by more than this angle are prohibited.  The
                  angle is specified in degrees.  Defaults to 90.

                - lock_boundary_vertices: bool.  Retain all vertices along mesh
                  surface boundaries, which can only occur at the boundary of
                  the volume.  Defaults to true.

        """
        super().__init__()
        self.token = make_random_token()
        if ts is not None and isinstance(data, ts.TensorStore):
            self.data = TensorStoreDataWrapper(data)
        else:
            self.data = DataWrapper(data)
        self.shape = self.data.shape
        rank = self.rank = len(self.shape)
        if dimensions is None:
            dimensions = CoordinateSpace(
                names=self.data.labels,
                units=self.data.units,
                scales=self.data.scales,
            )
        if rank != dimensions.rank:
            raise ValueError(
                "rank of data (%d) must match rank of coordinate space (%d)"
                % (rank, dimensions.rank)
            )
        if voxel_offset is None:
            voxel_offset = np.array(self.data.origin, dtype=np.int64)
        else:
            voxel_offset = np.array(voxel_offset, dtype=np.int64)
        if voxel_offset.shape != (rank,):
            raise ValueError("voxel_offset must have shape of (%d,)" % (rank,))
        self.voxel_offset = voxel_offset
        self.dimensions = dimensions
        self.data_type = np.dtype(self.data.dtype).name
        if self.data_type == "float64":
            self.data_type = "float32"
        self.encoding = encoding
        if volume_type is None:
            if self.rank == 3 and (
                self.data_type == "uint16"
                or self.data_type == "uint32"
                or self.data_type == "uint64"
            ):
                volume_type = "segmentation"
            else:
                volume_type = "image"
        self.volume_type = volume_type

        self._mesh_generator = None
        self._mesh_generator_pending = None
        self._mesh_generator_lock = threading.Condition()
        self._mesh_options = mesh_options.copy() if mesh_options is not None else dict()

        self.max_voxels_per_chunk_log2 = max_voxels_per_chunk_log2

        self.downsampling_layout = downsampling
        if chunk_layout is None:
            if downsampling == "2d":
                chunk_layout = "flat"
            else:
                chunk_layout = "isotropic"
        self.chunk_layout = chunk_layout

        self.max_downsampling = max_downsampling
        self.max_downsampled_size = max_downsampled_size
        self.max_downsampling_scales = max_downsampling_scales

    def info(self):
        info = dict(
            dataType=self.data_type,
            encoding=self.encoding,
            generation=self.change_count,
            coordinateSpace=self.dimensions.to_json(),
            shape=self.shape,
            volumeType=self.volume_type,
            voxelOffset=self.voxel_offset,
            chunkLayout=self.chunk_layout,
            downsamplingLayout=self.downsampling_layout,
            maxDownsampling=(
                None if math.isinf(self.max_downsampling) else self.max_downsampling
            ),
            maxDownsampledSize=(
                None
                if math.isinf(self.max_downsampled_size)
                else self.max_downsampled_size
            ),
            maxDownsamplingScales=(
                None
                if math.isinf(self.max_downsampling_scales)
                else self.max_downsampling_scales
            ),
        )
        if self.max_voxels_per_chunk_log2 is not None:
            info["maxVoxelsPerChunkLog2"] = self.max_voxels_per_chunk_log2

        return info

    def get_encoded_subvolume(self, data_format, start, end, scale_key):
        rank = self.rank
        if len(start) != rank or len(end) != rank:
            raise ValueError("Invalid request")
        downsample_factor = np.array(scale_key.split(","), dtype=np.int64)
        if (
            len(downsample_factor) != rank
            or np.any(downsample_factor < 1)
            or np.any(downsample_factor > self.max_downsampling)
            or np.prod(downsample_factor) > self.max_downsampling
        ):
            raise ValueError("Invalid downsampling factor.")
        downsampled_shape = np.asarray(
            np.ceil(self.shape / downsample_factor), dtype=np.int64
        )
        if np.any(end < start) or np.any(start < 0) or np.any(end > downsampled_shape):
            raise ValueError("Out of bounds data request.")

        orig_data_shape = self.data.shape
        indexing_expr = tuple(
            np.s_[
                start[i] * downsample_factor[i] : min(
                    orig_data_shape[i], end[i] * downsample_factor[i]
                )
            ]
            for i in range(rank)
        )
        subvol = np.array(self.data[indexing_expr], copy=False)
        if subvol.dtype == "float64":
            subvol = np.asarray(subvol, dtype=np.float32)

        if np.any(downsample_factor != 1):
            if self.volume_type == "image":
                subvol = downsample.downsample_with_averaging(subvol, downsample_factor)
            else:
                subvol = downsample.downsample_with_striding(subvol, downsample_factor)
        content_type = "application/octet-stream"
        if data_format == "jpeg":
            data = encode_jpeg(subvol)
            content_type = "image/jpeg"
        elif data_format == "npz":
            data = encode_npz(subvol)
        elif data_format == "raw":
            data = encode_raw(subvol)
        else:
            raise ValueError("Invalid data format requested.")
        return data, content_type

    def get_object_mesh(self, object_id):
        mesh_generator = self._get_mesh_generator()
        data = mesh_generator.get_mesh(object_id)
        if data is None:
            raise InvalidObjectIdForMesh()
        return data

    def _get_mesh_generator(self):
        if self._mesh_generator is not None:
            return self._mesh_generator
        while True:
            with self._mesh_generator_lock:
                if self._mesh_generator is not None:
                    return self._mesh_generator
                if self._mesh_generator_pending is not None:
                    while self._mesh_generator is None:
                        self._mesh_generator_lock.wait()
                    if self._mesh_generator is not None:
                        return self._mesh_generator
                try:
                    from . import _neuroglancer
                except ImportError:
                    raise MeshImplementationNotAvailable()
                if not (
                    self.rank == 3
                    and (
                        self.data_type == "uint8"
                        or self.data_type == "uint16"
                        or self.data_type == "uint32"
                        or self.data_type == "uint64"
                    )
                ):
                    raise MeshesNotSupportedForVolume()
                pending_obj = object()
                self._mesh_generator_pending = pending_obj
            data = self.data
            new_mesh_generator = _neuroglancer.OnDemandObjectMeshGenerator(
                data.transpose(),
                self.dimensions.scales,
                np.zeros(3),
                **self._mesh_options,
            )
            with self._mesh_generator_lock:
                if self._mesh_generator_pending is not pending_obj:
                    continue
                self._mesh_generator = new_mesh_generator
                self._mesh_generator_pending = False
                self._mesh_generator_lock.notify_all()
            return new_mesh_generator

    def __deepcopy__(self, memo):
        """Since this type is immutable, we don't need to deepcopy it.

        Actually deep copying would intefere with the use of deepcopy by JsonObjectWrapper.
        """
        return self

    def invalidate(self):
        """Mark the data invalidated.  Clients will refetch the volume."""
        with self._mesh_generator_lock:
            self._mesh_generator_pending = None
            self._mesh_generator = None
        self._dispatch_changed_callbacks()


class DataWrapper:
    """Wraps data for LocalVolume."""

    def __init__(self, data):
        self._data = data
        self._default_labels = ["d%d" % d for d in range(self.rank)]
        self._default_origin = [0] * self.rank
        self._default_scales = [1] * self.rank
        self._default_units = [""] * self.rank

    def __getitem__(self, item):
        return self._data.__getitem__(item)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._data, attr)

    @property
    def labels(self):
        return self._default_labels

    @property
    def origin(self):
        return self._default_origin

    @property
    def rank(self):
        return len(self._data.shape)

    @property
    def scales(self):
        return self._default_scales

    @property
    def units(self):
        return self._default_units


class TensorStoreDataWrapper(DataWrapper):
    """Wraps TensorStore data for LocalVolume."""

    def __init__(self, data):
        super().__init__(data=data[ts.d[:].translate_to[0]])
        self.dtype = self._data.dtype.numpy_dtype
        self._origin_before_translation = data.origin

    def __getitem__(self, item):
        return self._data.__getitem__(item).read().result()

    @property
    def labels(self):
        labels = self._default_labels
        for i, label in enumerate(self._data.domain.labels):
            if any(label):
                labels[i] = label
        return labels

    @property
    def origin(self):
        return self._origin_before_translation

    @property
    def scales(self):
        scales = self._default_scales
        for i, dimension_units in enumerate(self._data.dimension_units):
            if dimension_units is not None and dimension_units.multiplier is not None:
                scales[i] = dimension_units.multiplier
        return scales

    @property
    def units(self):
        units = self._default_units
        for i, dimension_units in enumerate(self._data.dimension_units):
            if dimension_units is not None and dimension_units.base_unit is not None:
                units[i] = dimension_units.base_unit
        return units

    def transpose(self):
        return TensorStoreDataWrapper(self._data.T)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Facilities for converting JSON <-> Python objects"""

import collections
import copy
import inspect
import numbers
import threading
from collections.abc import (
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    ValuesView,
)
from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    TypeVar,
    cast,
    overload,
)

import numpy as np
import numpy.typing

from .json_utils import encode_json_for_repr

__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


def to_json(value: Any) -> Any:
    if isinstance(value, np.ndarray):
        return value.tolist()
    try:
        method = value.to_json
    except AttributeError:
        return value
    return method()


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class JsonObjectWrapper:
    supports_readonly = True
    supports_validation = True

    __slots__ = ("_json_data", "_cached_wrappers", "_lock", "_readonly")

    _json_data: Any
    _cached_wrappers: dict[str, Any]
    _lock: threading.RLock
    _readonly: bool

    def __init__(self, json_data=None, _readonly=False, **kwargs):
        if json_data is None:
            json_data = {}
        elif isinstance(json_data, type(self)):
            json_data = json_data.to_json()
        elif not isinstance(json_data, dict):
            raise TypeError
        object.__setattr__(self, "_json_data", json_data)
        object.__setattr__(self, "_cached_wrappers", dict())
        object.__setattr__(self, "_lock", threading.RLock())
        object.__setattr__(self, "_readonly", 1 if _readonly else False)
        for k in kwargs:
            setattr(self, k, kwargs[k])
        object.__setattr__(self, "_readonly", _readonly)

    def to_json(self):
        if self._readonly:
            return self._json_data
        with self._lock:
            r = self._json_data.copy()
            for k, (wrapper, _) in self._cached_wrappers.items():
                if wrapper is not None:
                    r[k] = to_json(wrapper)
                else:
                    r.pop(k, None)
            return r

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(self.to_json(), memo))

    def __eq__(self, other):
        return type(self) == type(other) and self.to_json() == other.to_json()

    def __repr__(self):
        return f"{type(self).__name__}({encode_json_for_repr(self.to_json())})"

    def _get_wrapped(self, key, wrapped_type):
        with self._lock:
            json_value = self._json_data.get(key)
            cached_value = self._cached_wrappers.get(key)
            if cached_value is not None and cached_value[1] is json_value:
                return cached_value[0]
            kwargs = dict()
            if self._readonly and hasattr(wrapped_type, "supports_readonly"):
                kwargs["_readonly"] = True
            wrapper = wrapped_type(json_value, **kwargs)
            self._cached_wrappers[key] = wrapper, json_value
            return wrapper

    def _set_wrapped(self, key, value, validator):
        if self._readonly is True:
            raise AttributeError
        value = validator(value)
        with self._lock:
            self._cached_wrappers[key] = (value, self._json_data.get(key))


_types_supporting_validation = frozenset([np.uint64, float, int])


def _normalize_validator(wrapped_type, validator):
    if validator is None:
        supports_validation = getattr(wrapped_type, "supports_validation", None)
        if (
            inspect.isroutine(wrapped_type)
            or supports_validation is not None
            or wrapped_type in _types_supporting_validation
        ):
            if inspect.isroutine(supports_validation):
                validator = supports_validation
            else:
                validator = wrapped_type
        else:

            def validator_func(x):
                if not isinstance(x, wrapped_type):
                    raise TypeError(wrapped_type, x)
                return x

            validator = validator_func
    return validator


def _map_type_annotation(target, source, callback):
    annotation = _get_type_annotation(source)
    if annotation is None:
        return
    new_annotation = callback(annotation)
    if new_annotation is None:
        return
    _set_type_annotation(target, new_annotation)


def _set_type_annotation(target, annotation):
    setattr(target, "_neuroglancer_annotation", annotation)


def _get_type_annotation(wrapped_type):
    annotation = getattr(wrapped_type, "_neuroglancer_annotation", None)
    if annotation is not None:
        return annotation
    if isinstance(wrapped_type, type):
        return wrapped_type
    return None


def wrapped_property(json_name, wrapped_type, validator=None, doc=None):
    validator = _normalize_validator(wrapped_type, validator)

    def fget(self):
        return self._get_wrapped(json_name, wrapped_type)

    annotation = _get_type_annotation(wrapped_type)
    if annotation is not None:
        fget.__annotations__ = {"return": annotation}

    def fset(self, value):
        return self._set_wrapped(json_name, value, validator)

    return property(
        fget=fget,
        fset=fset,
        doc=doc,
    )


def array_wrapper(dtype, shape=None):
    if shape is not None:
        if isinstance(shape, numbers.Number):
            shape = (shape,)
        else:
            shape = tuple(shape)
        shape_annotation = tuple[tuple(Literal[s] for s in shape)]
    else:
        shape_annotation = Any

    def wrapper(value, _readonly=False):
        value = np.array(value, dtype=dtype)
        if _readonly:
            value.setflags(write=False)
        if shape is not None:
            if len(shape) != len(value.shape) or any(
                expected_size is not None and expected_size != actual_size
                for expected_size, actual_size in zip(shape, value.shape)
            ):
                raise ValueError("expected shape", shape)
        return value

    wrapper.supports_readonly = True
    _set_type_annotation(
        wrapper,
        (
            np.ndarray[shape_annotation, dtype]
            if shape_annotation is not Any
            else np.typing.NDArray[dtype]
        ),
    )
    return wrapper


def optional(wrapper, default_value=None, validator=None):
    def modified_wrapper(value, **kwargs):
        if value is None:
            return default_value
        return wrapper(value, **kwargs)

    if hasattr(wrapper, "supports_readonly"):
        modified_wrapper.supports_readonly = True

    validator = _normalize_validator(wrapper, validator)

    def modified_validator(value, **kwargs):
        if value is None:
            return default_value
        return validator(value, **kwargs)

    modified_wrapper.supports_validation = modified_validator

    if default_value is None:
        _map_type_annotation(modified_wrapper, wrapper, lambda t: t | None)
    else:
        _map_type_annotation(modified_wrapper, wrapper, lambda t: t)

    return modified_wrapper


@export
class Map(Generic[K, V], JsonObjectWrapper):
    """Maps keys of type :py:param:`.K` to values of type :py:param:`.V`.

    Type parameters:
      K:
        Key type.
      V:
        Mapped value type.

    Group:
      json-containers
    """

    _key_validator: ClassVar[Callable[[Any], Any]]
    _value_validator: ClassVar[Callable[[Any], Any]]
    _value_type: type
    supports_validation = True
    __slots__ = ()

    def __init__(self, json_data=None, _readonly=False):
        if isinstance(json_data, Map):
            json_data = json_data.to_json()
        elif json_data is not None:
            new_map = {}
            key_validator = type(self)._key_validator
            value_validator = type(self)._value_validator
            for k, v in json_data.items():
                key_validator(k)
                value_validator(v)
                new_map[str(k)] = to_json(v)
            json_data = new_map
        super().__init__(json_data, _readonly=_readonly)

    def clear(self):
        """Clears the map."""
        with self._lock:
            self._cached_wrappers.clear()
            self._json_data.clear()

    def keys(self) -> KeysView[K]:
        """Returns a dynamic view of the keys in the map."""
        return _MapKeysView(self)

    def values(self) -> ValuesView[V]:
        """Returns a dynamic view of the values in the map."""
        return _MapValuesView(self)

    def items(self) -> ItemsView[K, V]:
        """Returns a dynamic view of the items in the map."""
        return _MapItemsView(self)

    @overload
    def get(self, key: K) -> V | None: ...

    @overload
    def get(self, key: K, default: V) -> V: ...

    @overload
    def get(self, key: K, default: T) -> V | T: ...

    def get(self, key: K, default=None):
        """Returns the mapped value, or the specified default."""
        key = str(key)  # type: ignore[assignment]
        with self._lock:
            if key in self._json_data:
                return self._get_wrapped(key, type(self)._value_type)
            return default

    def __len__(self) -> int:
        """Returns the number of entries in the map."""
        return len(self._json_data)

    def __contains__(self, key: K) -> bool:
        return str(key) in self._json_data

    def __getitem__(self, key: K) -> V:
        """Returns the mapped value associated with the specified key.

        Raises:
          KeyError: if the key is not present in the map.
        """
        str_key = str(key)
        with self._lock:
            if str_key not in self._json_data:
                raise KeyError(key)
            return self._get_wrapped(str_key, type(self)._value_type)

    def __setitem__(self, key: K, value: V):
        """Sets the specified key to the specified value."""
        str_key = str(key)
        with self._lock:
            self._set_wrapped(str_key, value, type(self)._value_validator)
            self._json_data[str_key] = None  # placeholder

    def __delitem__(self, key: K):
        """Deletes the entry with the specified key.

        Raises:
          KeyError: if the key is not present in the map.
        """
        if self._readonly:
            raise AttributeError
        str_key = str(key)
        with self._lock:
            del self._json_data[str_key]
            self._cached_wrappers.pop(str_key, None)

    def __iter__(self) -> Iterator[K]:
        key_validator = type(self)._key_validator
        for key in self._json_data:
            yield key_validator(key)

    @overload
    def pop(self, key: K, /) -> V: ...

    @overload
    def pop(self, key: K, default: V, /) -> V: ...

    @overload
    def pop(self, key: K, default: T, /) -> V | T: ...

    def pop(self, key: K, /, *args):
        """Removes and returns the mapped value associated with the specified key.

        Returns:
          The mapped value, or :py:param:`default` if :py:param:`key` is not
          specified and :py:param:`default` is specified.

        Raises:
          KeyError: if the key is not present and :py:param:`default` is not
            specified.
        """
        if self._readonly:
            raise AttributeError
        str_key = str(key)
        if len(args) > 1:
            raise ValueError("Expected at most one default argument")
        with self._lock:
            if str_key in self._json_data:
                value = self._get_wrapped(str_key, type(self)._value_type)
                del self._json_data[str_key]
                self._cached_wrappers.pop(str_key, None)
                return value
            if len(args) == 0:
                raise KeyError(key)
            return args[0]


class _MapKeysView(Generic[K], collections.abc.KeysView[K]):
    _mapping: Map[K, Any]
    _base_view: KeysView[str]
    _key_validator: Callable[[str], K]

    def __init__(self, map: Map[K, Any]):
        self._mapping = map
        self._base_view = map._json_data.keys()
        self._key_validator = type(map)._key_validator

    def __contains__(self, key) -> bool:
        return str(key) in self._base_view

    def __len__(self) -> int:
        return len(self._base_view)

    def __iter__(self) -> Iterator[K]:
        key_validator = self._key_validator
        for key in self._base_view:
            yield key_validator(key)


class _MapItemsView(Generic[K, V], collections.abc.ItemsView[K, V]):
    _mapping: Map[K, V]

    def __init__(self, map: Map[K, V]):
        self._mapping = map

    def __contains__(self, item) -> bool:
        key, value = item
        m = self._mapping
        return key in m and m[key] == value

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[tuple[K, V]]:
        m = self._mapping
        for key in m:
            yield (key, m[key])


class _MapValuesView(Generic[V], collections.abc.ValuesView[V]):
    _mapping: Map[Any, V]

    def __init__(self, map: Map[Any, V]):
        self._mapping = map

    def __contains__(self, value) -> bool:
        return any(value == x for x in self)

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[V]:
        m = self._mapping
        for key in m:
            yield m[key]


def typed_map(key_type, value_type, key_validator=None, value_validator=None):
    key_validator = _normalize_validator(key_type, key_validator)
    value_validator = _normalize_validator(value_type, value_validator)

    class _Map(Map):
        __slots__ = ()
        _key_validator = key_validator
        _value_validator = value_validator
        _value_type = value_type

    if (key_annotation := _get_type_annotation(key_type)) is not None and (
        value_annotation := _get_type_annotation(value_type)
    ) is not None:
        _set_type_annotation(Map, Map[key_annotation, value_annotation])

    return _Map


def typed_set(wrapped_type: Callable[[Any], T]):
    def wrapper(x, _readonly=False) -> set[T] | frozenset[T]:
        set_type = frozenset if _readonly else set
        kwargs: dict[str, Any] = dict()
        if hasattr(wrapped_type, "supports_readonly"):
            kwargs.update(_readonly=True)
        if x is None:
            return set_type()
        return set_type(wrapped_type(v, **kwargs) for v in x)

    wrapper.supports_readonly = True  # type: ignore[attr-defined]
    _map_type_annotation(
        wrapper,
        wrapped_type,
        lambda t: set[t],  # type: ignore[valid-type]
    )
    return wrapper


@export
class List(Generic[T]):
    """List of values of type :py:param:`.T`.

    Type parameters:
      T: Element type.

    Group:
      json-containers
    """

    supports_readonly = True
    supports_validation = True
    __slots__ = ("_readonly", "_data")
    _validator: ClassVar[Callable[[Any], Any]]

    _readonly: bool
    _data: list[T]

    def __init__(self, json_data=None, _readonly=False):
        if json_data is None:
            json_data = []
        if not isinstance(json_data, list | tuple | np.ndarray):
            raise ValueError
        self._readonly = _readonly
        validator = type(self)._validator
        self._data = [validator(x) for x in json_data]

    def __len__(self) -> int:
        """Returns the length of the list."""
        return len(self._data)

    def __getitem__(self, key: int) -> T:
        """Returns the element at the specified index."""
        return self._data[key]

    def __delitem__(self, key: int):
        """Removes the element at the specified index."""
        if self._readonly:
            raise AttributeError
        del self._data[key]

    @overload
    def __setitem__(self, key: int, value: T): ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[T]): ...

    def __setitem__(self, key: int | slice, value: T | Iterable[T]):
        """Assigns to the specified index or slice."""
        if self._readonly:
            raise AttributeError
        if isinstance(key, slice):
            values = [type(self)._validator(x) for x in cast(Iterable[T], value)]
            self._data[key] = values
        else:
            value = type(self)._validator(value)
            self._data[key] = cast(T, value)

    def __iter__(self) -> Iterator[T]:
        """Iterates over the values in the list."""
        return iter(self._data)

    def append(self, value: T):
        """Appends a value to the end of the list."""
        if self._readonly:
            raise AttributeError
        value = type(self)._validator(value)
        self._data.append(value)

    def extend(self, values: Iterable[T]):
        """Extends the list with the specified values."""
        for x in values:
            self.append(x)

    def insert(self, index: int, value: T):
        """Inserts the specified value at the specified index."""
        value = type(self)._validator(value)
        self._data.insert(index, value)

    def pop(self, index: int = -1) -> T:
        """Removes and returns the element at the specified index."""
        return self._data.pop(index)

    def to_json(self):
        """Returns the representation as a JSON array."""
        return [to_json(x) for x in self._data]

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(self.to_json(), memo))

    def __repr__(self):
        return encode_json_for_repr(self.to_json())


def typed_list(wrapped_type: Callable[[Any], T], validator=None) -> type[List[T]]:
    val = _normalize_validator(wrapped_type, validator)

    class _List(List):
        _validator = val

    _map_type_annotation(_List, wrapped_type, lambda t: List[t])  # type: ignore[valid-type]

    return _List


def number_or_string(value):
    if not isinstance(value, numbers.Real | str):
        raise TypeError
    return value


_set_type_annotation(number_or_string, numbers.Real | str)


def bool_or_string(value):
    if not isinstance(value, bool | str):
        raise TypeError
    return value


_set_type_annotation(bool_or_string, bool | str)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Geometric transforms.

.. currentmodule:: nitransforms

.. autosummary::
   :toctree: ../generated

   transform
"""
from . import linear, manip, nonlinear, surface
from .linear import Affine, LinearTransformsMapping
from .nonlinear import DenseFieldTransform
from .manip import TransformChain
from .resampling import apply

try:
    from ._version import __version__
except ModuleNotFoundError:
    __version__ = "0+unknown"

__packagename__ = "nitransforms"
__copyright__ = "Copyright (c) 2021 The NiPy developers"

__all__ = [
    "apply",
    "surface",
    "linear",
    "manip",
    "nonlinear",
    "Affine",
    "LinearTransformsMapping",
    "DenseFieldTransform",
    "TransformChain",
    "__copyright__",
    "__packagename__",
    "__version__",
]



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Linear transforms."""

import warnings
from collections import namedtuple
import numpy as np
from pathlib import Path

from nibabel.affines import from_matvec

# Avoids circular imports
try:
    from ._version import __version__
except ModuleNotFoundError:  # pragma: no cover
    __version__ = "0+unknown"

from .base import (
    ImageGrid,
    TransformBase,
    _as_homogeneous,
    EQUALITY_TOL,
)
from .io import get_linear_factory, TransformFileError

try:
    from .io.x5 import (
        X5Transform,
        X5Domain,
        to_filename as save_x5,
        from_filename as load_x5,
    )
    _X5_AVAILABLE = True
except ImportError:
    _X5_AVAILABLE = False


class Affine(TransformBase):
    """Represents linear transforms on image data."""

    __slots__ = ("_matrix", "_inverse")

    def __init__(self, matrix=None, reference=None):
        """
        Initialize a linear transform.

        Parameters
        ----------
        matrix : ndarray
            The coordinate transformation matrix **in physical
            coordinates**, mapping coordinates from *reference* space
            into *moving* space.
            This matrix should be provided in homogeneous coordinates.

        Examples
        --------
        >>> xfm = Affine(reference=test_dir / "someones_anatomy.nii.gz")
        >>> xfm.matrix  # doctest: +NORMALIZE_WHITESPACE
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        >>> xfm = Affine([[1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        >>> xfm.matrix  # doctest: +NORMALIZE_WHITESPACE
        array([[1, 0, 0, 4],
               [0, 1, 0, 0],
               [0, 0, 1, 0],
               [0, 0, 0, 1]])

        """
        super().__init__(reference=reference)
        self._matrix = np.eye(4)
        self._inverse = np.eye(4)

        if matrix is not None:
            matrix = np.array(matrix)
            if matrix.ndim != 2:
                raise TypeError("Affine should be 2D.")
            elif matrix.shape[0] != matrix.shape[1]:
                raise TypeError("Matrix is not square.")
            self._matrix = matrix

            if not np.allclose(self._matrix[3, :], (0, 0, 0, 1)):
                raise ValueError(
                    """The last row of a homogeneus matrix \
should be (0, 0, 0, 1), got %s."""
                    % self._matrix[3, :]
                )

            # Normalize last row
            self._matrix[3, :] = (0, 0, 0, 1)
            self._inverse = np.linalg.inv(self._matrix)

    def __repr__(self):
        """
        Change representation to the internal matrix.

        Example
        -------
        >>> Affine([
        ...     [1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]
        ... ])  # doctest: +NORMALIZE_WHITESPACE
        array([[1, 0, 0, 4],
               [0, 1, 0, 0],
               [0, 0, 1, 0],
               [0, 0, 0, 1]])

        """
        return repr(self.matrix)

    def __eq__(self, other):
        """
        Overload equals operator.

        Examples
        --------
        >>> xfm1 = Affine([[1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        >>> xfm2 = Affine(xfm1.matrix)
        >>> xfm1 == xfm2
        True
        >>> xfm1 == Affine()
        False
        >>> xfm1 == TransformBase()
        False

        """
        if not hasattr(other, "matrix"):
            return False

        _eq = np.allclose(self.matrix, other.matrix, rtol=EQUALITY_TOL)
        if _eq and self._reference != other._reference:
            warnings.warn("Affines are equal, but references do not match.")
        return _eq

    def __invert__(self):
        """
        Get the inverse of this transform.

        Example
        -------
        >>> matrix = [[1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
        >>> Affine(np.linalg.inv(matrix)) == ~Affine(matrix)
        True

        """
        return self.__class__(self._inverse)

    def __len__(self):
        """Enable using len()."""
        return 1 if self._matrix.ndim == 2 else len(self._matrix)

    def __matmul__(self, b):
        """
        Compose two Affines.

        Example
        -------
        >>> xfm1 = Affine([[1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        >>> xfm1 @ ~xfm1 == Affine()
        True

        >>> xfm1 = Affine([[1, 0, 0, 4], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        >>> xfm1 @ np.eye(4) == xfm1
        True

        """
        if not isinstance(b, self.__class__):
            _b = self.__class__(b)
        else:
            _b = b

        retval = self.__class__(self.matrix.dot(_b.matrix))
        if _b.reference:
            retval.reference = _b.reference
        return retval

    @property
    def matrix(self):
        """Access the internal representation of this affine."""
        return self._matrix

    @property
    def ndim(self):
        """Access the internal representation of this affine."""
        return self._matrix.ndim + 1

    @classmethod
    def from_filename(
        cls, filename, fmt=None, reference=None, moving=None, x5_position=0
    ):
        """Create an affine from a transform file."""

        if fmt and fmt.upper() == "X5":
            return from_x5(
                load_x5(filename), reference=reference, x5_position=x5_position
            )

        fmtlist = [fmt] if fmt is not None else ("itk", "lta", "afni", "fsl")

        if fmt is not None and not Path(filename).exists():
            if fmt != "fsl":
                raise FileNotFoundError(
                    f"[Errno 2] No such file or directory: '{filename}'"
                )
            elif not Path(f"{filename}.000").exists():
                raise FileNotFoundError(
                    f"[Errno 2] No such file or directory: '{filename}[.000]'"
                )

        is_array = cls != Affine
        errors = []
        for potential_fmt in fmtlist:
            if potential_fmt == "itk" and Path(filename).suffix == ".mat":
                is_array = False
                cls = Affine

            try:
                struct = get_linear_factory(
                    potential_fmt, is_array=is_array
                ).from_filename(filename)
            except (TransformFileError, FileNotFoundError) as err:
                errors.append((potential_fmt, err))
                continue

            matrix = struct.to_ras(reference=reference, moving=moving)
            return cls(matrix, reference=reference)

        raise TransformFileError(
            f"Could not open <{filename}> (formats tried: {', '.join(fmtlist)})."
        )

    @classmethod
    def from_matvec(cls, mat=None, vec=None, reference=None):
        """
        Create an affine from a matrix and translation pair.

        Example
        -------
        >>> Affine.from_matvec(vec=(4, 0, 0))  # doctest: +NORMALIZE_WHITESPACE
        array([[1., 0., 0., 4.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        """
        mat = mat if mat is not None else np.eye(3)
        vec = vec if vec is not None else np.zeros((3,))
        return cls(from_matvec(mat, vector=vec), reference=reference)

    def map(self, x, inverse=False):
        r"""
        Apply :math:`y = f(x)`.

        Parameters
        ----------
        x : N x D numpy.ndarray
            Input RAS+ coordinates (i.e., physical coordinates).
        inverse : bool
            If ``True``, apply the inverse transform :math:`x = f^{-1}(y)`.

        Returns
        -------
        y : N x D numpy.ndarray
            Transformed (mapped) RAS+ coordinates (i.e., physical coordinates).

        Examples
        --------
        >>> xfm = Affine([[1, 0, 0, 1], [0, 1, 0, 2], [0, 0, 1, 3], [0, 0, 0, 1]])
        >>> xfm.map((0,0,0))
        array([[1., 2., 3.]])

        >>> xfm.map((0,0,0), inverse=True)
        array([[-1., -2., -3.]])

        """
        affine = self._matrix
        coords = _as_homogeneous(x, dim=affine.shape[0] - 1).T
        if inverse is True:
            affine = self._inverse
        return affine.dot(coords).T[..., :-1]

    def to_filename(self, filename, fmt="X5", moving=None, x5_inverse=False):
        """Store the transform in the requested output format."""
        if fmt.upper() == "X5":
            return save_x5(filename, [self.to_x5(store_inverse=x5_inverse)])

        writer = get_linear_factory(
            fmt, is_array=isinstance(self, LinearTransformsMapping)
        )

        if fmt.lower() in ("itk", "ants", "elastix"):
            writer.from_ras(self.matrix).to_filename(filename)
        else:
            # Rest of the formats peek into moving and reference image grids
            writer.from_ras(
                self.matrix,
                reference=self.reference,
                moving=ImageGrid(moving) if moving is not None else self.reference,
            ).to_filename(filename)
        return filename

    def to_x5(self, store_inverse=False, metadata=None):
        """Return an :class:`~nitransforms.io.x5.X5Transform` representation."""
        if not _X5_AVAILABLE:
            raise ImportError("h5py")

        metadata = {"WrittenBy": f"NiTransforms {__version__}"} | (metadata or {})

        domain = None
        if (reference := self.reference) is not None:
            domain = X5Domain(
                grid=True,
                size=getattr(reference or {}, "shape", (0, 0, 0)),
                mapping=reference.affine,
                coordinates="cartesian",
            )
        kinds = tuple("space" for _ in range(self.ndim)) + ("vector",)
        return X5Transform(
            type="linear",
            subtype="affine",
            representation="matrix",
            metadata=metadata,
            transform=self.matrix,
            dimension_kinds=kinds,
            domain=domain,
            inverse=(~self).matrix if store_inverse else None,
            array_length=len(self),
        )


class LinearTransformsMapping(Affine):
    """Represents a series of linear transforms."""

    def __init__(self, transforms, reference=None):
        """
        Initialize a linear transform mapping.

        Parameters
        ----------
        transforms : :obj:`list`
            The inverse coordinate transformation matrix **in physical
            coordinates**, mapping coordinates from *reference* space
            into *moving* space.
            This matrix should be provided in homogeneous coordinates.

        Examples
        --------
        >>> xfm = LinearTransformsMapping([
        ...     [[1., 0, 0, 1.], [0, 1., 0, 2.], [0, 0, 1., 3.], [0, 0, 0, 1.]],
        ...     [[1., 0, 0, -1.], [0, 1., 0, -2.], [0, 0, 1., -3.], [0, 0, 0, 1.]],
        ... ])
        >>> xfm[0].matrix  # doctest: +NORMALIZE_WHITESPACE
        array([[1., 0., 0., 1.],
               [0., 1., 0., 2.],
               [0., 0., 1., 3.],
               [0., 0., 0., 1.]])
        >>> (~xfm)[0].matrix  # doctest: +NORMALIZE_WHITESPACE
        array([[ 1., 0., 0., -1.],
               [ 0., 1., 0., -2.],
               [ 0., 0., 1., -3.],
               [ 0., 0., 0.,  1.]])

        """
        super().__init__(reference=reference)

        self._matrix = np.stack(
            [
                (xfm if isinstance(xfm, Affine) else Affine(xfm)).matrix
                for xfm in transforms
            ],
            axis=0,
        )
        self._inverse = np.linalg.inv(self._matrix)

    def __getitem__(self, i):
        """Enable indexed access to the series of matrices."""
        return Affine(self.matrix[i, ...], reference=self._reference)

    def map(self, x, inverse=False):
        r"""
        Apply :math:`y = f(x)`.

        Parameters
        ----------
        x : N x D numpy.ndarray
            Input RAS+ coordinates (i.e., physical coordinates).
        inverse : bool
            If ``True``, apply the inverse transform :math:`x = f^{-1}(y)`.

        Returns
        -------
        y : N x D numpy.ndarray
            Transformed (mapped) RAS+ coordinates (i.e., physical coordinates).

        Examples
        --------
        >>> xfm = LinearTransformsMapping([
        ...     [[1., 0, 0, 1.], [0, 1., 0, 2.], [0, 0, 1., 3.], [0, 0, 0, 1.]],
        ...     [[1., 0, 0, -1.], [0, 1., 0, -2.], [0, 0, 1., -3.], [0, 0, 0, 1.]],
        ... ])
        >>> xfm.matrix
        array([[[ 1.,  0.,  0.,  1.],
                [ 0.,  1.,  0.,  2.],
                [ 0.,  0.,  1.,  3.],
                [ 0.,  0.,  0.,  1.]],
        <BLANKLINE>
               [[ 1.,  0.,  0., -1.],
                [ 0.,  1.,  0., -2.],
                [ 0.,  0.,  1., -3.],
                [ 0.,  0.,  0.,  1.]]])

        >>> y = xfm.map([(0, 0, 0), (-1, -1, -1), (1, 1, 1)])
        >>> y[0, :, :3]
        array([[1., 2., 3.],
               [0., 1., 2.],
               [2., 3., 4.]])

        >>> y = xfm.map([(0, 0, 0), (-1, -1, -1), (1, 1, 1)], inverse=True)
        >>> y[0, :, :3]
        array([[-1., -2., -3.],
               [-2., -3., -4.],
               [ 0., -1., -2.]])


        """
        affine = self.matrix
        coords = _as_homogeneous(x, dim=affine.shape[-1] - 1).T
        if inverse is True:
            affine = self._inverse
        return np.swapaxes(affine.dot(coords), 1, 2)


def load(filename, fmt=None, reference=None, moving=None):
    """
    Load a linear transform file.

    Examples
    --------
    >>> xfm = load(regress_dir / "affine-LAS.itk.tfm")
    >>> isinstance(xfm, Affine)
    True

    >>> xfm = load(regress_dir / "itktflist.tfm")
    >>> isinstance(xfm, LinearTransformsMapping)
    True

    """
    xfm = LinearTransformsMapping.from_filename(
        filename, fmt=fmt, reference=reference, moving=moving
    )

    if isinstance(xfm, LinearTransformsMapping) and len(xfm) == 1:
        xfm = xfm[0]

    return xfm


def from_x5(x5_list, reference=None, x5_position=0):
    """Create an affine from a list of :class:`~nitransforms.io.x5.X5Transform` objects."""

    if not _X5_AVAILABLE:
        raise ImportError("h5py")

    x5_xfm = x5_list[x5_position]
    Transform = Affine if x5_xfm.array_length == 1 else LinearTransformsMapping
    if (
        x5_xfm.domain and not x5_xfm.domain.grid and len(x5_xfm.domain.size) == 3
    ):  # pragma: no cover
        raise NotImplementedError("Only 3D regularly gridded domains are supported")
    elif x5_xfm.domain:
        # Override reference
        Domain = namedtuple("Domain", "affine shape")
        reference = Domain(x5_xfm.domain.mapping, x5_xfm.domain.size)

    return Transform(x5_xfm.transform, reference=reference)



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modification notice (2025) Yael Balbastre
#   Imports of modules that depend on tornado (server, viewer) were
#   removed to allow use of this module within pyscript.

from . import (  # noqa: I001
    segment_colors,  # noqa: F401
    # server,  # noqa: F401
    skeleton,  # noqa: F401
)
from .default_credentials_manager import set_boss_token  # noqa: F401
from .equivalence_map import EquivalenceMap  # noqa: F401
from .local_volume import LocalVolume  # noqa: F401
from .screenshot import ScreenshotSaver  # noqa: F401
# from .server import (
#     is_server_running,  # noqa: F401
#     set_server_bind_address,  # noqa: F401
#     set_static_content_source,  # noqa: F401
#     set_dev_server_content_source,  # noqa: F401
#     stop,  # noqa: F401
# )
from .url_state import parse_url, to_json_dump, to_url  # noqa: F401
# from .viewer import UnsynchronizedViewer, Viewer  # noqa: F401
from .viewer_config_state import (
    LayerSelectedValues,  # noqa: F401
    LayerSelectionState,  # noqa: F401
    PrefetchState,  # noqa: F401
    ScaleBarOptions,  # noqa: F401
    SegmentIdMapEntry,  # noqa: F401
)
from .viewer_state import (
    CoordinateSpace,  # noqa: F401
    DimensionScale,  # noqa: F401
    CoordinateArray,  # noqa: F401
    Tool,  # noqa: F401
    PlacePointTool,  # noqa: F401
    PlaceLineTool,  # noqa: F401
    PlaceBoundingBoxTool,  # noqa: F401
    PlaceEllipsoidTool,  # noqa: F401
    BlendTool,  # noqa: F401
    OpacityTool,  # noqa: F401
    VolumeRenderingTool,  # noqa: F401
    VolumeRenderingGainTool,  # noqa: F401
    VolumeRenderingDepthSamplesTool,  # noqa: F401
    CrossSectionRenderScaleTool,  # noqa: F401
    SelectedAlphaTool,  # noqa: F401
    NotSelectedAlphaTool,  # noqa: F401
    ObjectAlphaTool,  # noqa: F401
    HideSegmentZeroTool,  # noqa: F401
    HoverHighlightTool,  # noqa: F401
    BaseSegmentColoringTool,  # noqa: F401
    IgnoreNullVisibleSetTool,  # noqa: F401
    ColorSeedTool,  # noqa: F401
    SegmentDefaultColorTool,  # noqa: F401
    MeshRenderScaleTool,  # noqa: F401
    MeshSilhouetteRenderingTool,  # noqa: F401
    SaturationTool,  # noqa: F401
    SkeletonRenderingMode2dTool,  # noqa: F401
    SkeletonRenderingMode3dTool,  # noqa: F401
    SkeletonRenderingLineWidth2dTool,  # noqa: F401
    SkeletonRenderingLineWidth3dTool,  # noqa: F401
    ShaderControlTool,  # noqa: F401
    MergeSegmentsTool,  # noqa: F401
    SplitSegmentsTool,  # noqa: F401
    SelectSegmentsTool,  # noqa: F401
    DimensionTool,  # noqa: F401
    SidePanelLocation,  # noqa: F401
    SelectedLayerState,  # noqa: F401
    StatisticsDisplayState,  # noqa: F401
    LayerSidePanelState,  # noqa: F401
    LayerListPanelState,  # noqa: F401
    HelpPanelState,  # noqa: F401
    DimensionPlaybackVelocity,  # noqa: F401
    Layer,  # noqa: F401
    PointAnnotationLayer,  # noqa: F401
    CoordinateSpaceTransform,  # noqa: F401
    LayerDataSubsource,  # noqa: F401
    LayerDataSource,  # noqa: F401
    LayerDataSources,  # noqa: F401
    InvlerpParameters,  # noqa: F401
    TransferFunctionParameters,  # noqa: F401
    ImageLayer,  # noqa: F401
    SkeletonRenderingOptions,  # noqa: F401
    StarredSegments,  # noqa: F401
    VisibleSegments,  # noqa: F401
    SegmentationLayer,  # noqa: F401
    SingleMeshLayer,  # noqa: F401
    Annotation,  # noqa: F401
    PointAnnotation,  # noqa: F401
    LineAnnotation,  # noqa: F401
    AxisAlignedBoundingBoxAnnotation,  # noqa: F401
    EllipsoidAnnotation,  # noqa: F401
    AnnotationPropertySpec,  # noqa: F401
    AnnotationLayer,  # noqa: F401
    LocalAnnotationLayer,  # noqa: F401
    ManagedLayer,  # noqa: F401
    Layers,  # noqa: F401
    LinkedPosition,  # noqa: F401
    LinkedZoomFactor,  # noqa: F401
    LinkedDepthRange,  # noqa: F401
    LinkedOrientationState,  # noqa: F401
    CrossSection,  # noqa: F401
    CrossSectionMap,  # noqa: F401
    DataPanelLayout,  # noqa: F401
    StackLayout,  # noqa: F401
    row_layout,  # noqa: F401
    column_layout,  # noqa: F401
    LayerGroupViewer,  # noqa: F401
    ViewerState,  # noqa: F401
)



# @license
# Copyright 2019-2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing a Neuroglancer coordinate space."""

import re
from collections.abc import Sequence
from typing import Any, NamedTuple

import numpy as np
import numpy.typing

__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


si_prefixes = {
    "Y": 24,
    "Z": 21,
    "E": 18,
    "P": 15,
    "T": 12,
    "G": 9,
    "M": 6,
    "k": 3,
    "h": 2,
    "": 0,
    "c": -2,
    "m": -3,
    "u": -6,
    "µ": -6,
    "n": -9,
    "p": -12,
    "f": -15,
    "a": -18,
    "z": -21,
    "y": -24,
}

si_units = ["m", "s", "rad/s", "Hz"]

si_units_with_prefixes = {
    f"{prefix}{unit}": (unit, exponent)
    for (prefix, exponent) in si_prefixes.items()
    for unit in si_units
}

si_units_with_prefixes[""] = ("", 0)


def parse_unit(scale, unit):
    unit, exponent = si_units_with_prefixes[unit]
    if exponent >= 0:
        return (scale * 10**exponent, unit)
    else:
        return (scale / 10 ** (-exponent), unit)


def parse_unit_and_scale(
    unit_and_scale: str, coefficient: float = 1.0
) -> tuple[float, str]:
    if unit_and_scale == "":
        return (coefficient, "")
    m = re.fullmatch(
        r"^((?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?)?([µa-zA-Z]+)?$", unit_and_scale
    )
    if m is None:
        raise ValueError("Invalid unit", unit_and_scale)
    scale_str = m.group(1)
    if scale_str is None:
        scale = 1.0
    else:
        scale = float(scale_str)

    scale *= coefficient
    unit = ""
    unit_str = m.group(2)
    if unit_str is not None:
        unit, exponent = si_units_with_prefixes[unit_str]
        if exponent >= 0:
            scale *= 10**exponent
        else:
            scale /= 10 ** (-exponent)
    return (scale, unit)


@export
class CoordinateArray:
    __slots__ = "_data"

    def __init__(self, json_data=None, labels=None, coordinates=None, mappings=None):
        if mappings is None:
            mappings = dict()
        else:
            mappings = dict(mappings)
        if labels is not None:
            if coordinates is None:
                coordinates = range(len(labels))
            for coordinate, label in zip(coordinates, labels):
                mappings[coordinate] = label
        if json_data is not None:
            if (
                not isinstance(json_data, dict)
                or "coordinates" not in json_data
                or "labels" not in json_data
            ):
                raise ValueError(
                    'Expected object with "coordinates" and "labels" properties'
                )
            coordinates = json_data["coordinates"]
            labels = json_data["labels"]
            for coordinate, label in zip(coordinates, labels):
                mappings[coordinate] = label
        self._data = mappings

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return repr(self._data)

    def __str__(self):
        return str(self._data)

    def __eq__(self, other):
        if not isinstance(other, CoordinateArray):
            return False
        return self._data == other._data

    def __getitem__(self, k):
        if isinstance(k, str):
            for other_k, other_v in self._data.items():
                if other_k == k:
                    return other_v
            raise KeyError(f"label not found: {k!r}")
        return self._data[k]

    def to_json(self):
        return dict(
            coordinates=list(self._data.keys()), labels=list(self._data.values())
        )


@export
class DimensionScale(NamedTuple):
    scale: float = 1
    """Voxel scaling along the dimension."""

    unit: str = ""
    """Units of `.scale`."""

    coordinate_array: CoordinateArray | None = None
    """Coordinate array for the dimension."""

    @staticmethod
    def from_json(json):
        if isinstance(json, DimensionScale):
            return json
        if isinstance(json, list):
            if len(json) != 2:
                raise ValueError(f"Expected [scale, unit], but received: {json!r}")
            scale = json[0]
            unit = json[1]
            coordinate_array = None
        else:
            scale = None
            unit = None
            coordinate_array = CoordinateArray(json_data=json)
        return DimensionScale(scale=scale, unit=unit, coordinate_array=coordinate_array)


@export
class CoordinateSpace:
    __slots__ = ("names", "scales", "units", "coordinate_arrays")

    names: tuple[str, ...]
    """Name of each dimension.

    Length is equal to `.rank`.
    """

    scales: np.typing.NDArray[np.float64]
    """Physical scale coefficient for `.unit` for each dimension.

    Length is equal to `.rank`.
    """

    units: tuple[str, ...]
    """Physical unit for each dimension.

    Length is equal to `.rank`.
    """

    coordinate_arrays: tuple[CoordinateArray | None, ...]
    """Coordinate array for each dimension.

    Length is equal to `.rank`.
    """

    def __init__(
        self,
        json: Any = None,
        names: Sequence[str] | None = None,
        scales: Sequence[float] | None = None,
        units: str | Sequence[str] | None = None,
        coordinate_arrays: Sequence[CoordinateArray | None] | None = None,
    ):
        """
        Constructs a coordinate space.

        Args:
          json: JSON representation.
          names: Dimension names (e.g., ['x', 'y', 'z']).
          scales: Voxel spacing along each dimension.
          units: Units of the values in :py:param:`.scales`.
          coordinate_arrays: Coordinate arrays associated with each dimension.
        """
        if json is None:
            if names is not None:
                names_tuple = tuple(names)
                rank = len(names_tuple)
                self.names = names_tuple
                scales_array: np.typing.NDArray[np.float64]
                if scales is None:
                    scales_array = np.ones(rank, dtype=np.float64)
                else:
                    scales_array = np.array(scales, dtype=np.float64)
                if units is None:
                    units = ""
                if isinstance(units, str):
                    units = tuple(units for _ in names_tuple)
                scales_and_units = tuple(
                    parse_unit_and_scale(unit, scale)
                    for scale, unit in zip(scales_array, units)
                )
                scales_array = np.array(
                    [s[0] for s in scales_and_units], dtype=np.float64
                )
                units = tuple(s[1] for s in scales_and_units)
                if coordinate_arrays is None:
                    coordinate_arrays = tuple(None for _ in units)
                else:
                    coordinate_arrays = tuple(coordinate_arrays)
                self.units = units
                self.scales = scales_array
                self.coordinate_arrays = coordinate_arrays
            else:
                self.names = ()
                self.scales = np.zeros(0, dtype=np.float64)
                self.units = ()
                self.coordinate_arrays = ()
        else:
            if not isinstance(json, dict):
                raise TypeError
            self.names = tuple(json.keys())
            values = tuple(DimensionScale.from_json(v) for v in json.values())
            self.scales = np.array([v.scale for v in values], dtype=np.float64)
            self.units = tuple(v.unit for v in values)
            self.coordinate_arrays = tuple(v.coordinate_array for v in values)
        self.scales.setflags(write=False)

    @property
    def rank(self) -> int:
        """Number of dimensions."""
        return len(self.names)

    def __getitem__(self, i):
        if isinstance(i, str):
            idx = self.names.index(i)
            return DimensionScale(
                scale=self.scales[idx],
                unit=self.units[idx],
                coordinate_array=self.coordinate_arrays[idx],
            )
        if isinstance(i, slice):
            idxs = range(self.rank)[i]
            return [
                DimensionScale(
                    scale=self.scales[j],
                    unit=self.units[j],
                    coordinate_array=self.coordinate_arrays[j],
                )
                for j in idxs
            ]
        return DimensionScale(
            scale=self.scales[i],
            unit=self.units[i],
            coordinate_array=self.coordinate_arrays[i],
        )

    def __repr__(self):
        return f"CoordinateSpace({self.to_json()!r})"

    def to_json(self):
        d = {}
        for name, scale, unit, coordinate_array in zip(
            self.names, self.scales, self.units, self.coordinate_arrays
        ):
            if coordinate_array is None:
                d[name] = [scale, unit]
            else:
                d[name] = coordinate_array.to_json()
        return d



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import re
import urllib.parse

from . import viewer_state
from .json_utils import json_encoder_default
from .json_wrappers import to_json

SINGLE_QUOTE_STRING_PATTERN = "('(?:[^'\\\\]|(?:\\\\.))*')"
DOUBLE_QUOTE_STRING_PATTERN = '("(?:[^"\\\\]|(?:\\\\.))*")'
SINGLE_OR_DOUBLE_QUOTE_STRING_PATTERN = (
    SINGLE_QUOTE_STRING_PATTERN + "|" + DOUBLE_QUOTE_STRING_PATTERN
)
DOUBLE_OR_SINGLE_QUOTE_STRING_PATTERN = (
    DOUBLE_QUOTE_STRING_PATTERN + "|" + SINGLE_QUOTE_STRING_PATTERN
)


DOUBLE_QUOTE_PATTERN = '^((?:[^"\'\\\\]|(?:\\\\.))*)"'
SINGLE_QUOTE_PATTERN = "^((?:[^\"'\\\\]|(?:\\\\.))*)'"


def _convert_string_literal(x, quote_initial, quote_replace, quote_search):
    if len(x) >= 2 and x[0] == quote_initial and x[-1] == quote_initial:
        inner = x[1:-1]
        s = quote_replace
        while inner:
            m = re.search(quote_search, inner)
            if m is None:
                s += inner
                break
            s += m.group(1)
            s += "\\"
            s += quote_replace
            inner = inner[m.end() :]
        s += quote_replace
        return s
    return x


def _convert_json_helper(x, desired_comma_char, desired_quote_char):
    comma_search = "[&_,]"
    if desired_quote_char == '"':
        quote_initial = "'"
        quote_search = DOUBLE_QUOTE_PATTERN
        string_literal_pattern = SINGLE_OR_DOUBLE_QUOTE_STRING_PATTERN
    else:
        quote_initial = '"'
        quote_search = SINGLE_QUOTE_PATTERN
        string_literal_pattern = DOUBLE_OR_SINGLE_QUOTE_STRING_PATTERN
    s = ""
    while x:
        m = re.search(string_literal_pattern, x)
        if m is None:
            before = x
            x = ""
            replacement = ""
        else:
            before = x[: m.start()]
            x = x[m.end() :]
            original_string = m.group(1)
            if original_string is not None:
                replacement = _convert_string_literal(
                    original_string, quote_initial, desired_quote_char, quote_search
                )
            else:
                replacement = m.group(2)
        s += re.sub(comma_search, desired_comma_char, before)
        s += replacement
    return s


def url_safe_to_json(x):
    return _convert_json_helper(x, ",", '"')


def json_to_url_safe(x):
    return _convert_json_helper(x, "_", "'")


def url_fragment_to_json(fragment_value):
    unquoted = urllib.parse.unquote(fragment_value)
    if unquoted.startswith("!"):
        unquoted = unquoted[1:]
    return url_safe_to_json(unquoted)


def parse_url_fragment(fragment_value) -> viewer_state.ViewerState:
    """Parses a Neuroglancer state from a URL fragment.

    Group:
      viewer-state-url
    """
    json_string = url_fragment_to_json(fragment_value)
    return viewer_state.ViewerState(json.loads(json_string))


def parse_url(url: str) -> viewer_state.ViewerState:
    """Parses a Neuroglancer state from a URL.

    Group:
      viewer-state-url
    """
    result = urllib.parse.urlparse(url)
    return parse_url_fragment(result.fragment)


def to_url_fragment(state: viewer_state.ViewerState):
    """Encodes a viewer state as a URL fragment.

    Group:
      viewer-state-url
    """
    json_string = json.dumps(
        to_json(state), separators=(",", ":"), default=json_encoder_default
    )
    return urllib.parse.quote(json_string, safe="~@#$&()*!+=:;,.?/'")


default_neuroglancer_url = "https://neuroglancer-demo.appspot.com"


def to_url(state: viewer_state.ViewerState, prefix=default_neuroglancer_url):
    """Encodes a viewer state as a URL.

    Group:
      viewer-state-url
    """
    return f"{prefix}#!{to_url_fragment(state)}"


def to_json_dump(state, indent=None, separators=None):
    """Returns the JSON-encoded text representation of the viewer state object.

    Group:
      viewer-state-url
    """
    return json.dumps(
        to_json(state),
        separators=separators,
        indent=indent,
        default=json_encoder_default,
    )



"""Utility to deal with optional dependencies."""
# stdlib
import importlib
from types import ModuleType


def _try_import(
    path: str,
    keys: str | list[str] | None = None,
    _as: bool = False
) -> ModuleType | object | None | tuple[ModuleType | object | None, ...]:
    """Try to import from a module.

    Example
    -------
    ```
    # import os.path
    os = try_import('os.path')

    # from os import path
    path = try_import('os', 'path')
    ```

    Parameters
    ----------
    path : str
        Path to module or variable in a module
    keys : str or list[str], optional
        Keys to load from the module
    _as : bool, defualt=True
        If False, perform recursive assignement as in
        >> # equivalent to: `import pack.sub.mod`
        >> pack = try_import('pack.sub.mod', _as=False)
        Else, it will look like a renamed import:
        >> # equivalent to: `import pack.sub.mod as my_mod`
        >> my_mod = try_import('pack.sub.mod', _as=True)


    Returns
    -------
    loaded_stuff : module or object or tuple
        A tuple is returned if `keys` is a list.
        Return None if import fails.

    """
    if path is None:
        return None

    def fail(keys: str | list[str] | None) -> None | list[None]:
        if keys is None or isinstance(keys, str):
            return None
        else:
            keys = list(keys)
            return [None]*len(keys)

    def try_import_module(path: str) -> ModuleType | None:
        try:
            return importlib.import_module(path)
        except (ImportError, ModuleNotFoundError):
            return None

    # check if the base package exists
    pack = path.split('.')[0]
    try:
        __import__(pack)
    except (ImportError, ModuleNotFoundError):
        return fail(keys)

    if _as:
        # import a module
        module = try_import_module(path)
        if not module:
            return fail(keys)
        # optional: extract attributes
        if keys is not None:
            if isinstance(keys, str):
                return getattr(module, keys)
            else:
                return tuple(getattr(module, key) for key in keys)
        return module
    else:
        # recursive import
        path = path.split('.')
        mod0 = try_import_module(path[0])
        if not mod0:
            return fail(keys)
        cursor = mod0
        for i in range(1, len(path)):
            mod1 = try_import_module('.'.join(path[:i+1]))
            if not mod1:
                return fail(keys)
            setattr(cursor, path[i], mod1)
            cursor = getattr(cursor, path[i])
        return mod0


def try_import(path: str, fallback: str | None = None) -> ModuleType | None:
    """Try to import a module.

    Example
    -------
    ```
    # import os
    os = try_import('os')

    # import os.path
    os = try_import('os.path')
    ```

    Parameters
    ----------
    path : str
        Path to module.
    fallaback: str, optional
        Path to fallback module

    Returns
    -------
    loaded_module : module | None
        Return `None` if import fails.

    """
    return _try_import(path) or _try_import(fallback)


def try_import_as(path: str, fallback: str | None = None) -> ModuleType | None:
    """Try to import a module and alias it.

    Example
    -------
    ```
    # import os.path as op
    op = try_import_as('os.path')
    ```

    Parameters
    ----------
    path : str
        Path to module.
    fallaback: str, optional
        Path to fallback module

    Returns
    -------
    loaded_module : module | None
        Return `None` if import fails.

    """
    return _try_import(path, _as=True) or _try_import(fallback, _as=True)


def try_from_import(
    path: str, keys: str | list[str], fallback: str | None = None,
) -> ModuleType | object | None | tuple[ModuleType | object | None, ...]:
    """Try to import attributes from a module.

    Example
    -------
    ```
    # from os import path
    path = try_from_import('os', 'path')

    # from os.path import join, splitext
    join, splitext = try_from_import('os.path', ['join', 'splitext'])

    # from os.path import join as J, splitext as S
    J, S = try_from_import('os.path', ['join', 'splitext'])
    ```

    Parameters
    ----------
    path : str
        Path to module.
    keys : str | list[str]
        Attributes to import from module
    fallaback: str, optional
        Path to fallback module

    Returns
    -------
    loaded_module : module | object | None | list[module | object | None]
        Return `None` if import fails.
        Return a list of objects if `keys` is a list.

    """
    return (
        _try_import(path, keys, _as=True) or
        _try_import(fallback, keys, _as=True)
    )
